(*
    Tux Commander - UEngines - Basic engines (local fs, ftp, archives...)
    Copyright (C) 2004 JASSER Software by Tomas Bzatek <tbzatek@users.sourceforge.net>
    Check for updates on tuxcmd.sourceforge.net

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)
unit UEngines;

interface

uses glib2, gdk2, Classes, Libc;


const ERRException = -1;
      ERRNoAccess = -2;
      ERRCreateLink = -3;
      ERRCopyMove = -4;
      ERRRemove = -5;
      ERRMkDIr = -6;

      omRead = 0;
      omWrite = 1;
      omAppend = 2;

      ConfDefaultDirCreationMask = 755;
      
type
  PDataItem = ^TDataItem;
  TDataItem = record
    AName, LnkPointTo: PChar;
    ColumnData: array[0..9] of PChar;
    Size: gint64;
    UpDir: boolean;
    Mode, UID, GID: Cardinal;
    IsDir, IsLnk, IsBlk, IsChr, IsFIFO, IsSock, Selected, IsDotFile: boolean;
    ModifyTime: TDateTime;
    Icon: Pointer;
    ItemColor: PGdkColor;
  end;

  PDataItemSL = ^TDataItemSL;
  TDataItemSL = record
    Stage1: boolean;
    AName, LnkPointTo, ADestination: PChar;
    Size: gint64;
    Mode, UID, GID: Cardinal;
    IsDir, IsLnk, ForceMove, IsOnRO, IsExecutable: boolean;
    ModifyTime: TDateTime;
    Level: word;
    atime, mtime: __time_t;
  end;

  TEngineProgressFunc = function (Sender: Pointer; BytesDone: Int64): boolean; cdecl;  //  Return False to break the copy process
  TEngineErrorFunc = function (Sender: Pointer; ErrorType, ErrorNum: integer; FileName: string): boolean; cdecl;  //  Return
  TEngineFileDes = pointer;

  TPanelEngine = class
  private
    BreakProcessingKind: integer;
  public
    ParentEngine: TPanelEngine;
    LastHighlightItem, SavePath: string;
    constructor Create;
    destructor Destroy; override;
//    function GetListing(var List: PGSList): integer; virtual; abstract;    //  Returns errorcode
    function GetListing(var List: TList; const AddDotFiles: boolean): integer; overload; virtual; abstract;    //  Returns errorcode
    function GetListing(var List: TList; const AddDotFiles: boolean; APath: string): integer; overload; virtual; abstract;    //  Returns errorcode
    function ChangeDir(const NewPath: string; const ShowProgress: boolean = True): integer; virtual; abstract;    //  Returns errorcode
    function ExplicitChDir(const NewPath: string): integer; virtual; abstract;    //  Returns errorcode
    function GetFileSystemSize: Int64; overload; virtual; abstract;
    function GetFileSystemSize(const APath: string): Int64; overload; virtual; abstract;
    function GetFileSystemFree: Int64; overload; virtual; abstract;
    function GetFileSystemFree(const APath: string): Int64; overload; virtual; abstract;
    function MakeDir(const NewDir: string): integer; virtual; abstract;    //  Returns errorcode
    function GetDirSize(APath: string): Int64; virtual; abstract;    //  Returns size or 0 if fails
    function Remove(APath: string): integer; virtual; abstract;    //  Returns errorcode
    procedure FillDirFiles(APath: string; List: TList; ALevel: word); virtual; abstract;
    function GetFileInfoSL(APath: string): PDataItemSL; virtual; abstract;
    function FileExists(const FileName: string; const Use_lstat: boolean = False): Boolean; virtual; abstract;
    function DirectoryExists(const FileName: string; const Use_lstat: boolean = False): Boolean; virtual; abstract;
    function MakeSymLink(const NewFileName, PointTo: string): integer; virtual; abstract;    //  Returns errorcode
    function Chmod(const FileName: string; const Mode: integer): integer; virtual; abstract;    //  Returns errorcode
    function Chown(const FileName: string; const UID, GID: integer): integer; virtual; abstract;    //  Returns errorcode
    procedure BreakProcessing(ProcessingKind: integer); virtual; abstract;
    function RenameFile(SourceFile, DestFile: string): integer; virtual; abstract;    //  Returns errorcode
    function ChangeTimes(APath: string; mtime, atime: __time_t): integer; virtual; abstract;    //  Returns errorcode
    procedure GetFileSystemInfo(const APath: string; var FSSize, FSFree: Int64; var FSName: string); virtual; abstract;
    function OpenFile(const APath: string; Mode: integer; var Error: integer): TEngineFileDes; virtual; abstract;  //  Returns filedescriptor
    function ReadFile(const FileDescriptor: TEngineFileDes; Buffer: Pointer; ABlockSize: integer; var Error: integer): integer; virtual; abstract;  //  Returns number of bytes read
    function WriteFile(const FileDescriptor: TEngineFileDes; Buffer: Pointer; BytesCount: integer; var Error: integer): integer; virtual; abstract;  //  Returns number of bytes written
    function CloseFile(const FileDescriptor: TEngineFileDes): integer; virtual; abstract;  //  Returns errorcode
    function FileSeek(const FileDescriptor: TEngineFileDes; const AbsoluteOffset: Int64; var Error: integer): Int64; virtual; abstract;  //  Returns errorcode
    function IsOnROMedium(const FileName: string): boolean; virtual; abstract;
    function FileCanRun(const FileName: string): boolean; virtual; abstract;
    function GetPath: string; virtual; abstract;
    procedure SetPath(Value: string); virtual; abstract;
    function GetPrefix: string; virtual; abstract;
    function Login(Username, Password: string): integer; virtual; abstract;

    //  Copy-related routines
    function GetBlockSize: guint32; virtual; abstract;
    procedure SetBlockSize(Value: guint32); virtual; abstract;
    function CopyFileIn(Sender: Pointer; SourceFile, DestFile: string; ProgressFunc: TEngineProgressFunc; ErrorFunc: TEngineErrorFunc; Append: boolean): boolean; virtual; abstract;  //  returns True if file is successfully copied
    function CopyFileOut(Sender: Pointer; SourceFile, DestFile: string; ProgressFunc: TEngineProgressFunc; ErrorFunc: TEngineErrorFunc; Append: boolean): boolean; virtual; abstract;  //  returns True if file is successfully copied
    function IsOnSameFS(const Path1, Path2: string): boolean; virtual; abstract;
    function TwoSameFiles(const Path1, Path2: string): boolean; virtual; abstract;
  published
    property Path: string read GetPath write SetPath;
    property BlockSize: guint32 read GetBlockSize write SetBlockSize;
  end;

  TLocalTreeEngine = class(TPanelEngine)
  private
    FPath: string;
    FBlockSize: guint32;
  public
    constructor Create;
    destructor Destroy; override;
//    function GetListing(var List: PGSList): integer; override;
    function GetListing(var List: TList; const AddDotFiles: boolean): integer; override;
    function GetListing(var List: TList; const AddDotFiles: boolean; APath: string): integer; override;
    function ChangeDir(const NewPath: string; const ShowProgress: boolean = True): integer; override;
    function ExplicitChDir(const NewPath: string): integer; override;
    function GetFileSystemSize: Int64; override;
    function GetFileSystemSize(const APath: string): Int64; override;
    function GetFileSystemFree: Int64; override;
    function GetFileSystemFree(const APath: string): Int64; override;
    function MakeDir(const NewDir: string): integer; override;
    function GetDirSize(APath: string): Int64; override;
    function Remove(APath: string): integer; override;
    procedure FillDirFiles(APath: string; List: TList; ALevel: word); override;
    function GetFileInfoSL(APath: string): PDataItemSL; override;
    function FileExists(const FileName: string; const Use_lstat: boolean = False): Boolean; override;
    function DirectoryExists(const FileName: string; const Use_lstat: boolean = False): Boolean; override;
    function MakeSymLink(const NewFileName, PointTo: string): integer; override;
    function Chmod(const FileName: string; const Mode: integer): integer; override;
    function Chown(const FileName: string; const UID, GID: integer): integer; override;
    procedure BreakProcessing(ProcessingKind: integer); override;
    function RenameFile(SourceFile, DestFile: string): integer; override;
    function ChangeTimes(APath: string; mtime, atime: __time_t): integer; override;
    procedure GetFileSystemInfo(const APath: string; var FSSize, FSFree: Int64; var FSName: string); override;
    function OpenFile(const APath: string; Mode: integer; var Error: integer): TEngineFileDes;  override;
    function ReadFile(const FileDescriptor: TEngineFileDes; Buffer: Pointer; ABlockSize: integer; var Error: integer): integer; override;
    function WriteFile(const FileDescriptor: TEngineFileDes; Buffer: Pointer; BytesCount: integer; var Error: integer): integer; override;
    function CloseFile(const FileDescriptor: TEngineFileDes): integer; override;
    function FileSeek(const FileDescriptor: TEngineFileDes; const AbsoluteOffset: Int64; var Error: integer): Int64; override;
    function IsOnROMedium(const FileName: string): boolean; override;
    function FileCanRun(const FileName: string): boolean; override;
    function GetPath: string; override;
    procedure SetPath(Value: string); override;
    function GetPrefix: string; override;
    function Login(Username, Password: string): integer; override;

    function GetBlockSize: guint32; override;
    procedure SetBlockSize(Value: guint32); override;
    function CopyFileIn(Sender: Pointer; SourceFile, DestFile: string; ProgressFunc: TEngineProgressFunc; ErrorFunc: TEngineErrorFunc; Append: boolean): boolean; override;
    function CopyFileOut(Sender: Pointer; SourceFile, DestFile: string; ProgressFunc: TEngineProgressFunc; ErrorFunc: TEngineErrorFunc; Append: boolean): boolean; override;
    function CopyFile(Sender: Pointer; SourceFile, DestFile: string; ProgressFunc: TEngineProgressFunc; ErrorFunc: TEngineErrorFunc; Append: boolean): boolean;
    function IsOnSameFS(const Path1, Path2: string): boolean; override;
    function TwoSameFiles(const Path1, Path2: string): boolean; override;
  published
    property Path;
    property BlockSize;
  end;


procedure FreeDataItem(DataItem: PDataItemSL); overload;
procedure FreeDataItem(DataItem: PDataItem); overload;


implementation

uses SysUtils, UCoreUtils, UConfig;

(********************************************************************************************************************************)
constructor TPanelEngine.Create;
begin
  inherited Create;
  BreakProcessingKind := 0;
  ParentEngine := nil;  //  By default it is a top-level engine (local)
  LastHighlightItem := '';
end;

destructor TPanelEngine.Destroy;
begin
  inherited Destroy;
end;

(********************************************************************************************************************************)
constructor TLocalTreeEngine.Create;
begin
  inherited Create;
  FPath := '/';
  FBlockSize := 65536;
end;

destructor TLocalTreeEngine.Destroy;
begin
  inherited Destroy;
end;

function TLocalTreeEngine.GetPath: string;
begin
  Result := FPath;
end;

procedure TLocalTreeEngine.SetPath(Value: string);
begin
  if Value <> FPath then begin
    FPath := Value;
  end;
end;

function TLocalTreeEngine.GetBlockSize: guint32;
begin
  Result := FBlockSize;
end;

procedure TLocalTreeEngine.SetBlockSize(Value: guint32);
begin
  if Value <> FBlockSize then begin
    FBlockSize := Value;
  end;
end;

function TLocalTreeEngine.GetListing(var List: {PGSList} TList; const AddDotFiles: boolean; APath: string): integer;
var Item : PDataItem;
    Handle : PDirectoryStream;
    DirEnt : PDirent64;
    Buf : PChar;
    StatBuf : TStatBuf64;
    i: integer;
    LnkBuf : array[0..1000] of char;
begin
  Result := 0;
  try
//  List := nil;
  APath := IncludeTrailingPathDelimiter(APath);
  if __chdir(PChar(APath)) <> 0 then begin
    Result := errno;
    Exit;
  end;
  Handle := OpenDir(PChar(APath));
  if not Assigned(Handle) then begin
    Result := ERRNoAccess;
    Exit;
  end;
  repeat
    DirEnt := readdir64(Handle);
    if Assigned(DirEnt) and Assigned(PChar(@DirEnt^.d_name[0])) then begin
      Buf := Pchar(@DirEnt^.d_name[0]);
      if (Buf <> '.') and (Buf <> '..') and (DirEnt^.d_name[0] <> #0) and
         (AddDotFiles or (Length(Buf) = 1) or ((Length(Buf) > 1) and (not ((Buf[0] = '.') and (Buf[1] <> '.'))))) then
      begin
        New(Item);
        with Item^ do begin
          AName := nil;
          LnkPointTo := nil;
          for i := 0 to Length(ColumnData) - 1 do ColumnData[i] := nil;
          GetMem(AName, Length(Buf) + 1);
          StrCopy(AName, Buf);
          if __lxstat64(_STAT_VER, Buf, StatBuf) <> 0 then begin
            DebugMsg(['*** Error reading file stat (lstat): ', strerror(errno)]);
            Continue;
          end;  
          Mode := StatBuf.st_mode;
          IsDotFile := (Length(Buf) > 1) and (Buf[0] = '.') and (Buf[1] <> '.');
          IsDir := __S_ISTYPE(StatBuf.st_mode, __S_IFDIR);
          IsLnk := __S_ISTYPE(StatBuf.st_mode, __S_IFLNK);
          IsBlk := __S_ISTYPE(StatBuf.st_mode, __S_IFBLK);
          IsChr := __S_ISTYPE(StatBuf.st_mode, __S_IFCHR);
          IsFIFO := __S_ISTYPE(StatBuf.st_mode, __S_IFIFO);
          IsSock := __S_ISTYPE(StatBuf.st_mode, __S_IFSOCK);
          ModifyTime := UnixTimeToTDateTime(StatBuf.st_mtime);
          if StatBuf.st_uid = 4294967295 then UID := getuid
                                         else UID := StatBuf.st_uid;
          if StatBuf.st_gid = 4294967295 then GID := getgid
                                         else GID := StatBuf.st_gid;
          UpDir := False;
          Selected := False;
          if IsLnk then begin
            i := readlink(PChar(APath + String(Buf)), LnkBuf, SizeOf(LnkBuf));
            if i > 0 then begin
              LnkBuf[i] := #0;
              GetMem(LnkPointTo, i + 1);
              StrLCopy(LnkPointTo, @LnkBuf[0], i);
            end;

            if __xstat64(_STAT_VER, Buf, StatBuf) = 0 then begin
              IsDir := __S_ISTYPE(StatBuf.st_mode, __S_IFDIR);
              Mode := StatBuf.st_mode;
            end;
          end;
          if not IsDir then Size := StatBuf.st_size
                       else Size := -1;
//          List := g_slist_append(List, Item);
          List.Add(Item);
        end;
      end;
    end;
  until DirEnt = nil;
  closedir(Handle);
  except
    on E: Exception do begin
      Result := ERRException;
      DebugMsg(['*** TLocalTreeEngine.GetListing(APath=', APath, ') -Exception: ', E.Message]);
      Exit;
    end;
  end;
end;

function TLocalTreeEngine.GetListing(var List: {PGSList} TList; const AddDotFiles: boolean): integer;
begin
  Result := GetListing(List, AddDotFiles, FPath);
end;

function TLocalTreeEngine.ChangeDir(const NewPath: string; const ShowProgress: boolean = True): integer;
var APath: string;
    Handle : PDirectoryStream;
begin
  try
    APath := IncludeTrailingPathDelimiter(NewPath);
    if __chdir(PChar(APath)) <> 0 then begin
      Result := errno;
      Exit;
    end;
    Handle := opendir(PChar(APath));
    if not Assigned(Handle) then begin
      Result := ERRNoAccess;
      Exit;
    end;
{    if not Assigned(readdir(Handle)) then begin
      Result := ERRNoAccess;
      Exit;
    end; }
    if closedir(Handle) <> 0 then begin
      Result := ERRNoAccess;
      Exit;
    end;
    Result := 0; 
  except
    on E: Exception do begin
      Result := ERRException;
      DebugMsg(['*** TLocalTreeEngine.ChangeDir(APath=', APath, ') -Exception: ', E.Message]);
      Exit;
    end;
  end;
end;

(********************************************************************************************************************************)
function TLocalTreeEngine.ExplicitChDir(const NewPath: string): integer;
begin
  Result := __chdir(PChar(NewPath));
  if Result <> 0 then Result := errno;
end;

(********************************************************************************************************************************)
function TLocalTreeEngine.GetFileSystemSize(const APath: string): Int64;
var Stat: TStatFS64;
begin
  Result := 0;
  if statfs64(PChar(APath), Stat) <> 0 then Exit;
  Result := Stat.f_bsize * Stat.f_blocks;
end;

function TLocalTreeEngine.GetFileSystemSize: Int64;
begin
  Result := GetFileSystemSize(FPath);
end;

(********************************************************************************************************************************)
function TLocalTreeEngine.GetFileSystemFree(const APath: string): Int64;
var Stat: TStatFS64;
begin
  Result := 0;
  if statfs64(PChar(APath), Stat) <> 0 then Exit;
  Result := Stat.f_bsize * Stat.f_bavail;
end;

function TLocalTreeEngine.GetFileSystemFree: Int64;
begin
  Result := GetFileSystemFree(FPath);
end;

(********************************************************************************************************************************)
function TLocalTreeEngine.MakeDir(const NewDir: string): integer;
begin
  Result := __mkdir(PChar(NewDir), OctalToAttr(ConfDefaultDirCreationMask));
  if Result <> 0 then try
    if Self.DirectoryExists(NewDir) or (not ForceDirectories(NewDir)) then Result := errno;
  except
    Result := -1;
  end;
end;

(********************************************************************************************************************************)

function TLocalTreeEngine.GetDirSize(APath: string): Int64;

  function InternalGetDirSize(APath: string): Int64;
  var Handle : PDirectoryStream;
      DirEnt : PDirent;
      StatBuf : TStatBuf64;
  begin
    Result := 0;
    try
      if BreakProcessingKind = 1 then Exit;
      APath := IncludeTrailingPathDelimiter(APath);
      if __chdir(PChar(APath)) <> 0 then begin
        Result := 0;
        Exit;
      end;
      Handle := OpenDir(PChar(APath));
      if not Assigned(Handle) then begin
        Result := 0;
        Exit;
      end;
      repeat
        DirEnt := readdir(Handle);
        if Assigned(DirEnt) and Assigned(PChar(@DirEnt^.d_name[0])) and (PChar(@DirEnt^.d_name[0]) <> '.') and
           (PChar(@DirEnt^.d_name[0]) <> '..') and (DirEnt^.d_name[0] <> #0) then
        begin
          if __lxstat64(_STAT_VER, PChar(@DirEnt^.d_name[0]), StatBuf) <> 0 then Continue;
          if __S_ISTYPE(StatBuf.st_mode, __S_IFDIR) then begin
            Inc(Result, InternalGetDirSize(APath + String(PChar(@DirEnt^.d_name[0]))));
            __chdir(PChar(APath));
          end else Inc(Result, StatBuf.st_size);
        end;
      until DirEnt = nil;
      CloseDir(Handle);
    except
      Result := 0;
    end;
  end;

begin
  try
    BreakProcessingKind := 0;
    Result := InternalGetDirSize(APath);
  finally
    BreakProcessingKind := 0;
  end;
end;

(********************************************************************************************************************************)
function TLocalTreeEngine.Remove(APath: string): integer;
begin
  APath := ExcludeTrailingPathDelimiter(APath);
  Result := Libc.remove(PChar(APath));
  if Result <> 0 then Result := errno;
end;

(********************************************************************************************************************************)
procedure TLocalTreeEngine.FillDirFiles(APath: string; List: TList; ALevel: word);
var Handle : PDirectoryStream;
    DirEnt : PDirent;
    StatBuf : TStatBuf64;
    Item: PDataItemSL;
    i: integer;
    LnkBuf : array[0..1000] of char;
    FilesList: TList;


  procedure AddEntry(FPath: string; AddCurrDirStage, AStage1: boolean);
  begin
    FPath := ExcludeTrailingPathDelimiter(FPath);
    if __lxstat64(_STAT_VER, PChar(FPath), StatBuf) <> 0 then begin
      DebugMsg(['*** Error reading file stat AddEntry(lstat): ', strerror(errno)]);
      Exit;
    end;
    New(Item);
    with Item^ do begin
      AName := nil;
      LnkPointTo := nil;
      ADestination := nil;
      Stage1 := AStage1;
      GetMem(AName, Length(FPath) + 1);
      StrCopy(AName, PChar(FPath));
      Size := StatBuf.st_size;
      Mode := StatBuf.st_mode;
      IsDir := __S_ISTYPE(StatBuf.st_mode, __S_IFDIR);
      IsLnk := __S_ISTYPE(StatBuf.st_mode, __S_IFLNK);
      IsExecutable := AddCurrDirStage or (StatBuf.st_mode and S_IXUSR = S_IXUSR);
      IsOnRO := IsOnROMedium(FPath);
      ForceMove := False;
      if StatBuf.st_uid = 4294967295 then UID := getuid
                                     else UID := StatBuf.st_uid;
      if StatBuf.st_gid = 4294967295 then GID := getgid
                                     else GID := StatBuf.st_gid;
      atime := StatBuf.st_atime;
      mtime := StatBuf.st_mtime;
      if IsLnk and AddCurrDirStage then DebugMsg(['*** Assertion failed AddEntry: Item^.IsLnk = True']);
      if IsLnk and (not AddCurrDirStage) then begin
        i := readlink(PChar(APath + String(PChar(@DirEnt^.d_name[0]))), LnkBuf, SizeOf(LnkBuf));
        if i > 0 then begin
          LnkBuf[i] := #0;
          GetMem(LnkPointTo, i + 1);
          StrLCopy(LnkPointTo, @LnkBuf[0], i);
        end;
      end;
      ModifyTime := UnixTimeToTDateTime(StatBuf.st_mtime);
      Level := ALevel + Ord(not AddCurrDirStage);
    end;
    if AddCurrDirStage then List.Add(Item)
                       else FilesList.Add(Item);
  end;

begin
  if not Assigned(List) then Exit;
  try
    AddEntry(APath, True, True);
    FilesList := TList.Create;
    APath := IncludeTrailingPathDelimiter(APath);
    if __chdir(PChar(APath)) <> 0 then Exit;
    Handle := OpenDir(PChar(APath));
    if Assigned(Handle) then
    repeat
      DirEnt := readdir(Handle);
      if Assigned(DirEnt) and Assigned(PChar(@DirEnt^.d_name[0])) and (PChar(@DirEnt^.d_name[0]) <> '.') and (PChar(@DirEnt^.d_name[0]) <> '..') then begin
        if __lxstat64(_STAT_VER, PChar(@DirEnt^.d_name[0]), StatBuf) <> 0 then Continue;
        if __S_ISTYPE(StatBuf.st_mode, __S_IFDIR) then begin
          FillDirFiles(APath + String(PChar(@DirEnt^.d_name[0])), List, ALevel + 1);
          __chdir(PChar(APath));
        end else AddEntry(APath + String(PChar(@DirEnt^.d_name[0])), False, True);
      end;
    until DirEnt = nil;
    CloseDir(Handle);
    if FilesList.Count > 0 then
      for i := 0 to FilesList.Count - 1 do
        List.Add(FilesList[i]);
    FilesList.Free;
    AddEntry(APath, True, False);
  except
    on E: Exception do DebugMsg(['*** TLocalTreeEngine.FillDirFiles(APath=', APath, ', Level=', ALevel, ') -Exception: ', E.Message]);
  end;
end;

(********************************************************************************************************************************)
function TLocalTreeEngine.GetFileInfoSL(APath: string): PDataItemSL;
var StatBuf : TStatBuf64;
    i : integer;
    LnkBuf : array[0..1000] of char;
begin
  Result := nil;
  if __lxstat64(_STAT_VER, PChar(APath), StatBuf) <> 0 then begin
    DebugMsg(['*** Error reading file stat GetFileInfoSL(lstat): ', strerror(errno)]);
    Exit;
  end;
  New(Result);
  with Result^ do begin
    AName := nil;
    LnkPointTo := nil;
    ADestination := nil;
    Stage1 := True;
    GetMem(AName, Length(APath) + 1);
    StrCopy(AName, PChar(APath));
    Size := StatBuf.st_size;
    Mode := StatBuf.st_mode;
    IsDir := __S_ISTYPE(StatBuf.st_mode, __S_IFDIR);
    IsLnk := __S_ISTYPE(StatBuf.st_mode, __S_IFLNK);
    IsExecutable := StatBuf.st_mode and S_IXUSR = S_IXUSR;
    IsOnRO := IsOnROMedium(APath);
    ForceMove := False;
    ModifyTime := UnixTimeToTDateTime(StatBuf.st_mtime);
    if StatBuf.st_uid = 4294967295 then UID := getuid
                                   else UID := StatBuf.st_uid;
    if StatBuf.st_gid = 4294967295 then GID := getgid
                                   else GID := StatBuf.st_gid;
    atime := StatBuf.st_atime;
    mtime := StatBuf.st_mtime;
    Level := 1;
    if IsLnk then begin
      i := readlink(PChar(APath), LnkBuf, SizeOf(LnkBuf));
      if i > 0 then begin
        LnkBuf[i] := #0;
        GetMem(LnkPointTo, i + 1);
        StrLCopy(LnkPointTo, @LnkBuf[0], i);
      end;
      if __xstat64(_STAT_VER, PChar(APath), StatBuf) = 0 then begin
        IsDir := __S_ISTYPE(StatBuf.st_mode, __S_IFDIR);
        Mode := StatBuf.st_mode;
      end;
    end;
  end;
end;

(********************************************************************************************************************************)
function TLocalTreeEngine.CopyFileIn(Sender: Pointer; SourceFile, DestFile: string; ProgressFunc: TEngineProgressFunc; ErrorFunc: TEngineErrorFunc; Append: boolean): boolean;
begin
  Result := CopyFile(Sender, SourceFile, DestFile, ProgressFunc, ErrorFunc, Append);
end;

function TLocalTreeEngine.CopyFileOut(Sender: Pointer; SourceFile, DestFile: string; ProgressFunc: TEngineProgressFunc; ErrorFunc: TEngineErrorFunc; Append: boolean): boolean;
begin
  Result := CopyFile(Sender, SourceFile, DestFile, ProgressFunc, ErrorFunc, Append);
end;

function TLocalTreeEngine.CopyFile(Sender: Pointer; SourceFile, DestFile: string; ProgressFunc: TEngineProgressFunc; ErrorFunc: TEngineErrorFunc; Append: boolean): boolean;
var fsrc, fdest: PIOFile;
    BytesDone, BytesRead: Int64;
    StatBuf: TStatBuf64;
    offset: off_t;


  //  Used due to sendfile bug while copying from NTFS and some 2.6.x kernels
  function OldCopyRoutine: boolean;
  var Buffer: Pointer;
      BytesWritten: Int64;
      Res: boolean;
  begin
    Result := False;
    Res := True;
    DebugMsg(['*** Using old copy function due to bug in sendfile']);
    Buffer := malloc(FBlockSize);
    if Buffer = nil then begin
      ErrorFunc(Sender, 1, errno, SourceFile);  //  Memory allocation failed
      Libc.free(Buffer);
      Exit;
    end;

    while feof(fsrc) = 0 do begin
      BytesRead := fread(Buffer, 1, FBlockSize, fsrc);
      if (BytesRead < FBlockSize) and (feof(fsrc) = 0) then begin
        Res := ErrorFunc(Sender, 6, errno, SourceFile);  //  Cannot read from source file
        Break;
      end;
      BytesWritten := fwrite(Buffer, 1, BytesRead, fdest);
      if BytesWritten < BytesRead then begin
        Res := ErrorFunc(Sender, 7, ferror(fdest), DestFile);  //  Cannot write to source file
        Break;
      end;
      Inc(BytesDone, BytesRead);
      if Assigned(ProgressFunc) and (not ProgressFunc(Sender, BytesDone)) then begin
        Res := False;
        Break;
      end;
    end;
    Libc.free(Buffer);
    Result := Res;
  end;

  function NewCopyRoutine: boolean;
  var Res: boolean;
  begin
    Res := True;
    repeat
//    DebugMsg(['Copy: offset = ', offset, ', BytesDone = ', BytesDone, ', ftell(fsrc) = ', ftell(fsrc)]);
      BytesRead := Libc.sendfile(fileno(fdest), fileno(fsrc), offset, FBlockSize);
      if BytesRead = -1 then begin
        if errno = EINVAL then begin
          Result := OldCopyRoutine;
          Exit;
        end else Res := ErrorFunc(Sender, 6, errno, SourceFile);  //  Cannot read from source file
        Break;
      end;
      Inc(BytesDone, BytesRead);
      if Assigned(ProgressFunc) and (not ProgressFunc(Sender, BytesDone)) then begin
        Res := False;
        Break;
      end;
    until BytesRead < FBlockSize;

    __lxstat64(_STAT_VER, PChar(SourceFile), StatBuf);
    if (BytesDone < StatBuf.st_size) and Res then
      Res := ErrorFunc(Sender, 6, errno, SourceFile);
    Result := Res;
  end;

begin
  Result := False;
  fsrc := fopen(PChar(SourceFile), 'r');
  if fsrc = nil then begin
    ErrorFunc(Sender, 2, errno, SourceFile);  //  Cannot open source file
    Exit;
  end;
  if Append then fdest := fopen(PChar(DestFile), 'a')
            else fdest := fopen(PChar(DestFile), 'w');
  if fdest = nil then begin
    fclose(fsrc);
    ErrorFunc(Sender, 3, errno, DestFile);  //  Cannot open target file
    Exit;
  end;

  BytesDone := 0;
  offset := 0;

  Result := NewCopyRoutine;

  if fclose(fdest) <> 0 then begin
    fclose(fsrc);
    ErrorFunc(Sender, 4, errno, DestFile);  //  Cannot close target file
    Exit;
  end;
  if fclose(fsrc) <> 0 then begin
    ErrorFunc(Sender, 5, errno, SourceFile);  //  Cannot close source file
    Exit;
  end;
end;

(********************************************************************************************************************************)
function TLocalTreeEngine.FileExists(const FileName: string; const Use_lstat: boolean = False): Boolean;
var st: TStatBuf;
begin
  if Use_lstat then Result := lstat(PChar(FileName), st) = 0
               else Result := stat(PChar(FileName), st) = 0;
end;

(********************************************************************************************************************************)
function TLocalTreeEngine.DirectoryExists(const FileName: string; const Use_lstat: boolean = False): Boolean;
var st: TStatBuf;
begin
  if Use_lstat then Result := lstat(PChar(FileName), st) = 0 else
   if stat(PChar(FileName), st) = 0
     then Result := S_ISDIR(st.st_mode)
     else Result := False;
end;

(********************************************************************************************************************************)
function TLocalTreeEngine.MakeSymLink(const NewFileName, PointTo: string): integer;
begin
  Result := Libc.symlink(PChar(PointTo), PChar(NewFileName));
  if Result <> 0 then Result := errno;
end;

(********************************************************************************************************************************)
function TLocalTreeEngine.Chmod(const FileName: string; const Mode: integer): integer;
begin
  Result := Libc.chmod(PChar(FileName), Mode);
  if Result <> 0 then Result := errno;
end;

(********************************************************************************************************************************)
function TLocalTreeEngine.Chown(const FileName: string; const UID, GID: integer): integer;
begin
  Result := Libc.chown(PChar(FileName), UID, GID);
  if Result <> 0 then Result := errno;
end;

(********************************************************************************************************************************)
procedure TLocalTreeEngine.BreakProcessing(ProcessingKind: integer);
begin
  BreakProcessingKind := ProcessingKind;
end;

(********************************************************************************************************************************)
function TLocalTreeEngine.IsOnSameFS(const Path1, Path2: string): boolean;
var FStat1, FStat2: TStatBuf64;
begin
  DebugMsg(['** TLocalTreeEngine.IsOnSameFS("', Path1, '", "', Path2, '")']);
  Result := False;    //  Default fallback result (forces copy + delete)
  FillChar(FStat1, SizeOf(FStat1), 0);
  if __lxstat64(_STAT_VER, PChar(Path1), FStat1) <> 0 then begin
    DebugMsg(['** TLocalTreeEngine.IsOnSameFS: Libc.stat(', Path1, ') error: ', strerror(errno)]);
    Exit;
  end;
  FillChar(FStat2, SizeOf(FStat2), 0);
  if __lxstat64(_STAT_VER, PChar(Path2), FStat2) <> 0 then begin
    DebugMsg(['** TLocalTreeEngine.IsOnSameFS: Libc.stat(', Path1, ') error: ', strerror(errno)]);
    Exit;
  end;
  Result := FStat1.st_dev = FStat2.st_dev;
  DebugMsg(['** TLocalTreeEngine.IsOnSameFS("', Path1, '", "', Path2, '")  Result = ', Result]);
end;

(********************************************************************************************************************************)
function TLocalTreeEngine.RenameFile(SourceFile, DestFile: string): integer;
begin
  Result := Libc.__rename(PChar(SourceFile), PChar(DestFile));
  if Result <> 0 then Result := errno;
end;

(********************************************************************************************************************************)
function TLocalTreeEngine.ChangeTimes(APath: string; mtime, atime: __time_t): integer;
var timebuf: utimbuf;
begin
  timebuf.actime := atime;
  timebuf.modtime := mtime;
  Result := utime(PChar(APath), @timebuf);
  if Result <> 0 then Result := errno;
end;

(********************************************************************************************************************************)
procedure TLocalTreeEngine.GetFileSystemInfo(const APath: string; var FSSize, FSFree: Int64; var FSName: string);
var Stat: TStatFS64;
    fd: PIOFile;
    mntent: PMountEntry;
    mntdev: PChar;
    FoundLength: integer;
    Buffer: array[0..31] of char;
begin
  FSSize := 0;
  FSFree := 0;
  FSName := '';
  if statfs64(PChar(APath), Stat) <> 0 then Exit;
  FSSize := Stat.f_bsize * Stat.f_blocks;
  FSFree := Stat.f_bsize * Stat.f_bavail;
  fd := setmntent(_PATH_MOUNTED, 'r');
  if fd = nil then Exit;
  //  Get mount name
  FoundLength := 0;
  mntdev := nil;
  mntent := getmntent(fd);
  while mntent <> nil do begin
    if (Pos(mntent^.mnt_dir, APath) = 1) and (Length(String(mntent^.mnt_dir)) > FoundLength) then begin
      FoundLength := Length(String(mntent^.mnt_dir));
      FSName := String(mntent^.mnt_dir);
      mntdev := mntent^.mnt_fsname;
    end;
    mntent := getmntent(fd);
  end;
  endmntent(fd);
  //  if it is CD-ROM, read ISO9660 label
  if Stat.f_type = $9660 then begin  { ISOFS_SUPER_MAGIC }
    if Assigned(mntdev) and (mntdev <> '') then begin
      fd := fopen(mntdev, 'r');
      if fd = nil then Exit;
      if fseek(fd, 32808, SEEK_SET) <> 0 then Exit;
      if fread(@Buffer[0], 1, 32, fd) <> 0 then FSName := Trim(String(Buffer));
      fclose(fd);
    end;
  end;
end;

(********************************************************************************************************************************)
function TLocalTreeEngine.OpenFile(const APath: string; Mode: integer; var Error: integer): TEngineFileDes;
var m: PChar;
begin
  Error := 0;
  case Mode of
    omRead: m := 'r';
    omWrite: m := 'w';
    omAppend: m := 'a';
    else m := 'r';
  end;
  Result := Libc.fopen64(PChar(APath), m);
  if Result = nil then Error := errno;
end;

(********************************************************************************************************************************)
function TLocalTreeEngine.ReadFile(const FileDescriptor: TEngineFileDes; Buffer: Pointer; ABlockSize: integer; var Error: integer): integer;
begin
  Error := 0;
  Result := Libc.fread(Buffer, 1, ABlockSize, FileDescriptor);
  if (Result = 0) and (feof(FileDescriptor) = 0) then Error := errno;
end;

(********************************************************************************************************************************)
function TLocalTreeEngine.WriteFile(const FileDescriptor: TEngineFileDes; Buffer: Pointer; BytesCount: integer; var Error: integer): integer;
begin
  Error := 0;
{  Result := Libc.__write(fileno(FileDescriptor), Buffer^, BytesCount);
  if Result < BytesCount then Error := errno; }
  Result := Libc.fwrite(Buffer, 1, BytesCount, FileDescriptor);
  if Result < BytesCount then Error := ferror(FileDescriptor); 
end;

(********************************************************************************************************************************)
function TLocalTreeEngine.CloseFile(const FileDescriptor: TEngineFileDes): integer;
begin
  Result := Libc.fclose(FileDescriptor);
  if Result <> 0 then Result := errno;
end;

(********************************************************************************************************************************)
function TLocalTreeEngine.FileSeek(const FileDescriptor: TEngineFileDes; const AbsoluteOffset: Int64; var Error: integer): Int64;
begin
  Error := 0;
  Result := Libc.fseeko64(FileDescriptor, AbsoluteOffset, SEEK_SET);
  if Result = -1 then Error := errno;
end;

(********************************************************************************************************************************)
function TLocalTreeEngine.IsOnROMedium(const FileName: string): boolean;
var Stat: TStatFS;
begin
  Result := False;
  if statfs(PChar(FileName), Stat) <> 0 then Exit;
  Result := (Stat.f_type = $9660) or { ISOFS_SUPER_MAGIC }
            (Stat.f_type = 1397118030); { NTFS }
end;

(********************************************************************************************************************************)
function TLocalTreeEngine.FileCanRun(const FileName: string): boolean;
begin
  Result := Libc.access(PChar(FileName), R_OK or X_OK) = 0;
end;

(********************************************************************************************************************************)
function TLocalTreeEngine.GetPrefix: string;
begin
  Result := '';
end;

(********************************************************************************************************************************)
function TLocalTreeEngine.Login(Username, Password: string): integer;
begin
  Result := 0;
end;

(********************************************************************************************************************************)
function TLocalTreeEngine.TwoSameFiles(const Path1, Path2: string): boolean;
var st1, st2: TStatBuf;
begin
  Result := False;
  if lstat(PChar(Path1), st1) <> 0 then Exit;
  if lstat(PChar(Path2), st2) <> 0 then Exit;
  DebugMsg(['*** TLocalTreeEngine.TwoSameFiles: ', st1.st_ino, ' ', st2.st_ino]);
  Result := st1.st_ino = st2.st_ino;
end;


(********************************************************************************************************************************)
(********************************************************************************************************************************)
procedure FreeDataItem(DataItem: PDataItemSL);
begin
  if Assigned(DataItem) then begin
    with DataItem^ do begin
      if Assigned(AName) then Dispose(AName);
      if Assigned(LnkPointTo) then Dispose(LnkPointTo);
      if Assigned(ADestination) then Dispose(ADestination);
    end;
    Dispose(DataItem);
  end;
end;

procedure FreeDataItem(DataItem: PDataItem);
var i : integer;
begin
  if Assigned(DataItem) then begin
    with DataItem^ do begin
      if AName <> nil then Dispose(AName);
      if LnkPointTo <> nil then Dispose(LnkPointTo);
      for i := 0 to Length(ColumnData) - 1 do
//        if ConfColumnVisible[i + 1] then
          if ColumnData[i] <> nil then try Dispose(ColumnData[i]); except end;
    end;
    try Dispose(DataItem); except end;
  end;
end;

end.


