/* null plugin for Tux Commander
 * Copyright (C) 2005 JASSER Software by Tomas Bzatek <tbzatek@users.sourceforge.net>
 *  Check for updates on tuxcmd.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include "vfs_types.h"


//  Declaration of the global plugin object
typedef struct {
  TVFSLogFunc log_func;
  int list_item_id;
  char *curr_dir;
} TVFSGlobs;



//  Basic initialization functions
int VFSAllocNeeded()
{
  return sizeof(TVFSGlobs);
}

void VFSInit(TVFSGlobs *globs, TVFSLogFunc log_func)
{

  globs->log_func = log_func;
  globs->log_func("null plugin: VFSInit");
  globs->curr_dir = NULL;

}

void VFSDestroy(TVFSGlobs *globs)
{

  globs->log_func("null plugin: VFSDestroy");

}

int VFSVersion()
{
  return cVFSVersion;
}

TVFSInfo VFSGetInfo()
{

  static const TVFSInfo info =
  {
     "Null plugin",
     "null testing plugin",
     "Plugin API version 2",
     "Copyright © 2005 JASSER Software by Tomáš Bžatek"
  };

  return info;
}

char *VFSGetPrefix(TVFSGlobs *globs)
{
  return "null";
}

TVFSResult VFSOpen(TVFSGlobs *globs, char *sName)
{
 return cVFS_OK;
}

TVFSResult VFSClose(TVFSGlobs *globs)
{
 return cVFS_OK;
}

char *VFSGetPath(TVFSGlobs *globs)
{
  return globs->curr_dir;
}

u_int64_t VFSGetFileSystemFree(TVFSGlobs *globs, char *APath)
{
  return 0;
}

u_int64_t VFSGetFileSystemSize(TVFSGlobs *globs, char *APath)
{
  return 0;
}



////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////

TVFSResult VFSChangeDir(TVFSGlobs *globs, char *NewPath)
{
  if (globs->curr_dir != NULL) free(globs->curr_dir);
  globs->curr_dir = strdup(NewPath);
  return cVFS_OK;
}

int VFSLogin(TVFSGlobs *globs, char *user, char *pass)
{
  return cVFS_OK;
}



////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////

static char *items[] =
{
   "This is the first item",
   "This is the second item",
   "This is the third item",
   "This is the fourth item",
   "This is the fifth item",
   "This is the sixth item",
   "This is the seventh item",
   "This is the eight item",
   "This is the nineth item",
   "This is the tenth item",
   "This is the eleventh item",
   "This is the twelveth item",
   "This is the thirteenth item",
   "This is the fourteenth item",
   "This is the fifteenth item",
   "This is the sixteeth item",
   "This is the seventeenth item",
   "This is the eighteenth item",
};


TVFSResult VFSListFirst(TVFSGlobs *globs, char *sDir, TVFSItem *Item)
{
  globs->list_item_id = 0;

  Item->sFileName = strdup(items[globs->list_item_id]);
  Item->iSize = rand();
  Item->iMode = S_IRWXO + S_IRWXG + S_IRWXU;
  Item->sLinkTo = NULL;
  Item->ItemType = vDirectory;

  return cVFS_OK;
}

TVFSResult VFSListNext(TVFSGlobs *globs, char *sDir, TVFSItem *Item)
{
  globs->list_item_id++;

  if (globs->list_item_id >= (sizeof(items) / sizeof(items[0]))) {
   Item = NULL;
   return cVFS_Not_More_Files;
  }

  Item->sFileName = strdup(items[globs->list_item_id]);

  Item->iSize = rand();
  Item->iMode = S_IRWXO + S_IRWXG + S_IRWXU;
  Item->sLinkTo = NULL;

  if (globs->list_item_id < 4) Item->ItemType = vDirectory;
  else Item->ItemType = vRegular;

  return cVFS_OK;
}

TVFSResult VFSListClose(TVFSGlobs *globs)
{
  return cVFS_OK;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////
long VFSFileExists(TVFSGlobs *globs, const char *FileName, const long Use_lstat)
{
  return 0;
}

TVFSResult VFSFileInfo(TVFSGlobs *globs, char *AFileName, TVFSItem *Item)
{
  return cVFS_OK;
}

TVFSResult VFSMkDir(TVFSGlobs *globs, const char *sDirName)
{
  return cVFS_OK;
}

TVFSResult VFSRemove(TVFSGlobs *globs, const char *APath)
{
  return cVFS_OK;
}

TVFSResult VFSRename(TVFSGlobs *globs, const char *sSrcName, const char *sDstName)
{
  return cVFS_OK;
}

TVFSResult VFSMakeSymLink(TVFSGlobs *globs, const char *NewFileName, const char *PointTo)
{
  return cVFS_OK;
}

TVFSResult VFSChmod(TVFSGlobs *globs, const char *FileName, const uint Mode)
{
  return cVFS_OK;
}

TVFSResult VFSChown(TVFSGlobs *globs, const char *FileName, const uint UID, const uint GID)
{
  return cVFS_OK;
}

TVFSResult VFSChangeTimes(TVFSGlobs *globs, char *APath, long mtime, long atime)
{
  return cVFS_OK;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////

u_int64_t VFSGetDirSize(TVFSGlobs *globs, char *APath)
{
  return 0;
}


void VFSBreakGetDirSize(TVFSGlobs *globs)
{

}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////

TVFSFileDes VFSOpenFile(TVFSGlobs *globs, const char *APath, int Mode, int *Error)
{
  *Error = 0;
  return (TVFSFileDes)0;
}

TVFSResult VFSCloseFile(TVFSGlobs *globs, TVFSFileDes FileDescriptor)
{
  return cVFS_OK;
}

u_int64_t VFSFileSeek(TVFSGlobs *globs, TVFSFileDes FileDescriptor, u_int64_t AbsoluteOffset, int *Error)
{
  *Error = 0;
  return 0;
}

int VFSReadFile(TVFSGlobs *globs, TVFSFileDes FileDescriptor, void *Buffer, int ABlockSize, int *Error)
{
  *Error = 0;
  return 0;
}

int VFSWriteFile(TVFSGlobs *globs, TVFSFileDes FileDescriptor, void *Buffer, int BytesCount, int *Error)
{
  *Error = 0;
  return 0;
}

void VFSSetBlockSize(TVFSGlobs *globs, int Value)
{

}

int VFSIsOnSameFS(TVFSGlobs *globs, const char *Path1, const char *Path2)
{
  return 0;
}

int VFSTwoSameFiles(TVFSGlobs *globs, const char *Path1, const char *Path2)
{
  return 0;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////

TVFSResult VFSCopyOut(TVFSGlobs *globs, const char *sSrcName, const char *sDstName, TVFSCopyCallBackFunc pCallBackProgress, void *data, int Append)
{
  return cVFS_OK;
}

TVFSResult VFSCopyIn(TVFSGlobs *globs, const char *sSrcName, const char *sDstName, TVFSCopyCallBackFunc pCallBackProgress, void *data, int Append)
{
  return cVFS_OK;
}

////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////




